# copyright 2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

__docformat__ = "restructuredtext en"

from pyramid.httpexceptions import HTTPSeeOther
from urllib.parse import urlparse

from cubicweb.predicates import anonymous_user
from cubicweb.tags import a
from cubicweb.web.views.basetemplates import LogInTemplate, LoggedOutTemplate

from cubicweb_saml.utils import retrieve_url_from_client


class SAMLLogMixin:
    """ Mixin class used to manage log-related templates
    """

    def get_saml_uri_from_request(self, request):
        """ Retrieve the SAML URI based on receiving request

        Parameters
        ----------
        request : pyramid.request.Request
            Pyramid request used to retrieve SAML response

        Returns
        -------
        str
            SAML URI as string
        """

        return retrieve_url_from_client(self._cw.vreg.config, request)

    def on_activate_saml(self):
        """ Retrieve and redirect to SAML URI
        """

        uri = self.retrieve_saml_uri()

        if uri is not None:
            self.redirect_to(uri)

    def redirect_to(self, uri):
        """ Redirect user to a dedicated URI

        Parameters
        ----------
        uri : string
            The URI used to redirect the user after an action

        Raises
        ------
        HTTPSeeOther
            Redirect to the specify URI string parameter
        """

        # Show a link on webpage in case of the redirection is not allowed in
        # web browser
        self.w(a('Redirect to SAML IdP...', href=uri))

        raise HTTPSeeOther(location=uri)

    def retrieve_saml_uri(self):
        """ Retrieve the SAML URI if sources has been correctly defined

        Returns
        -------
        str or None
            SAML URI as string if sources are configured, None otherwise
        """

        sources = self._cw.vreg.config.read_sources_file()

        metadata_uri = sources.get('saml', {}).get('saml-metadata-uri', None)
        if metadata_uri is None:
            return None

        request = self._cw._request

        # Add fake login form data to current request to start SAML connection
        # correctly. The SAML Relay State is used to store the postlogin_path
        # URI where the user is redirect after an successful SAML login
        request.GET.update({
            '__login': 'saml-login',
            'postlogin_path': request.url,
        })

        return self.get_saml_uri_from_request(request)


class SAMLLogin(LogInTemplate, SAMLLogMixin):
    """ Redirect automatically to SAML IdP server

    This class override the default login page to always redirect to IdP server

    See Also
    --------
    cubicweb.web.views.basetemplates.LogInTemplate
    cubicweb_saml.views.SAMLLogMixin
    """

    __select__ = LogInTemplate.__select__ & anonymous_user()
    __regid__ = 'login'

    priority = 50

    def call(self):
        """ Call the current view

        See Also
        --------
        cubicweb.web.views.basetemplates.LogInOutTemplate.call
        """

        self.on_activate_saml()
        super(SAMLLogin, self).call()


class SAMLLogout(LoggedOutTemplate, SAMLLogMixin):
    """

    See Also
    --------
    cubicweb.web.views.basetemplates.LoggedOutTemplate
    cubicweb_saml.views.SAMLLogMixin
    """

    __select__ = LoggedOutTemplate.__select__ & anonymous_user()
    __regid__ = 'loggedout'

    priority = 50

    def call(self):
        """ Call the current view

        See Also
        --------
        cubicweb.web.views.basetemplates.LoggedOutTemplate.call
        """

        self.on_activate_saml()
        super(SAMLLogout, self).call()

    def get_saml_uri(self, request):
        """ Retrieve the SAML IdP server URL based on receiving request

        Parameters
        ----------
        request : pyramid.request.Request
            Pyramid request used to retrieve SAML response

        Returns
        -------
        str
            SAML IdP server URL as string

        See Also
        --------
        cubicweb_saml.views.SAMLLogMixin.get_saml_uri
        """

        saml_uri = super(SAMLLogout, self).get_saml_uri(request)

        # Use urlparse to only retrieve the beggining of the uri string
        uri = urlparse(saml_uri)

        return '%s://%s' % (uri.scheme, uri.netloc)


def registration_callback(vreg):
    """ Add SAML dedicated views inside the CubicWeb registry

    Parameters
    ----------
    vreg : cubicweb.cwvreg.CWRegistryStore
        CubicWeb registry storage instance
    """

    vreg.register_all(globals().values(), __name__)



# copyright 2019-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from mock import patch
from pyramid import testing

from cubicweb import ValidationError, NoResultError
from cubicweb.devtools import testlib

from cubicweb_saml.utils import (build_base_url,
                                 execute_rql,
                                 find_group_by_name,
                                 find_user_by_login,
                                 generate_hash,
                                 get_user,
                                 register_user,
                                 register_user_into_db)


class _SAMLCommonTC(testlib.CubicWebTC):

    def setUp(self):
        super(_SAMLCommonTC, self).setUp()

        self.request = testing.DummyRequest(post={'SAMLResponse': ''})
        self.request.registry['cubicweb.config'] = self.config
        self.request.registry['cubicweb.repository'] = self.repo


class SAMLUtilsTC(_SAMLCommonTC):

    def test_build_base_url(self):
        config = self.vreg.config

        self.assertEqual(
            build_base_url(config), 'http://testing.fr/cubicweb/')

        config['base-url'] = 'http://url.without.backslash'

        self.assertEqual(
            build_base_url(config), 'http://url.without.backslash/')

    def test_generate_hash(self):
        self.assertNotEqual(generate_hash('example'),
                            generate_hash('example'))

        self.assertIsNotNone(generate_hash('example', 'with', 'arguments'))


class SAMLDatabaseTC(_SAMLCommonTC):

    def test_execute_rql(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertIsNone(execute_rql(
                cnx,
                "Any U WHERE U is CWUser, U login %(login)s",
                login="CubicWeb"))

            rset = execute_rql(cnx, "Any U WHERE U is CWUser, U login 'admin'")
            self.assertIsInstance(rset, int)
            self.assertEqual(rset, 6)

    def test_find_user_by_login(self):
        with self.admin_access.repo_cnx() as cnx:
            user_eid = register_user_into_db(cnx, "test", "managers", "psswd")

            find_eid = find_user_by_login(cnx, "test")
            self.assertEqual(user_eid, find_eid)

    def test_find_group_by_name(self):
        with self.admin_access.repo_cnx() as cnx:
            self.assertIsNotNone(find_group_by_name(cnx, "managers"))
            self.assertIsNotNone(find_group_by_name(cnx, "users"))
            self.assertIsNone(find_group_by_name(cnx, "watchmen"))

    def test_register_an_user_into_db(self):
        with self.admin_access.repo_cnx() as cnx:
            with self.assertRaises(NoResultError):
                cnx.execute("Any U WHERE U is CWUser, U login 'test'").one()

            user_eid = register_user_into_db(cnx, "test", "managers", "psswd")
            self.assertIsInstance(user_eid, int)

            user = cnx.entity_from_eid(user_eid)
            self.assertEqual(user.login, "test")
            self.assertIn("managers", user.groups)


class SAMLResponseTC(_SAMLCommonTC):

    def test_get_an_unknown_user(self):
        with patch('cubicweb_saml.utils.retrieve_identity_from_client',
                   return_value=('unknown_user', {})):

            self.assertIsNone(get_user(self.request))

    def test_get_an_existing_user(self):
        with self.admin_access.repo_cnx() as cnx:
            user = self.create_user(cnx, 'saml_user')

            with patch('cubicweb_saml.utils.retrieve_identity_from_client',
                       return_value=('saml_user', {})):

                userid = get_user(self.request)

                self.assertEqual(userid, user.eid)

    def test_register_a_new_user(self):
        with patch('cubicweb_saml.utils.retrieve_identity_from_client',
                   return_value=('saml_user', {})):

            self.assertIsNone(get_user(self.request))

            userid = register_user(self.request)

            self.assertIsNotNone(userid)
            self.assertEqual(get_user(self.request), userid)

    def test_register_an_existing_user(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, 'saml_user')

            with patch('cubicweb_saml.utils.retrieve_identity_from_client',
                       return_value=('saml_user', {})):

                self.assertIsNotNone(get_user(self.request))

                with self.assertRaises(ValidationError):
                    register_user(self.request)


if __name__ == '__main__':
    from unittest import main
    main()



# -*- coding: utf-8 -*-
# pylint: disable=W0622
# copyright 2017-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


"""cubicweb-saml application packaging information"""


modname = 'cubicweb_saml'
distname = 'cubicweb-saml'

numversion = (0, 5, 0)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = 'SAML2 authentifier'
web = 'https://forge.extranet.logilab.fr/cubicweb/cubes/%s' % distname

__depends__ = {
    'cubicweb': ">= 3.30.0, < 3.34.0",
    'logilab-common': '>= 1.4.0',
    'pyramid': '>= 1.10.4',
    'pysaml2': '>= 5.0.0',
    'requests': '>= 2.22.0',
    'zope.interface': '>= 4.1.2',
}
__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
]



# copyright 2019-2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging

logger = logging.getLogger(__name__)


def includeme(config):
    """ Activate the SAML authenticate cube
    """
    from cubicweb_saml.pconfig import SAMLAuthenticationPolicy

    if config.registry.get('cubicweb.authpolicy') is None:
        raise ValueError("saml: the default cubicweb auth policy should be "
                         "available via the 'cubicweb.authpolicy' registry "
                         "config entry")

    cubicweb_sources = config.registry['cubicweb.config'].read_sources_file()

    if 'saml' not in cubicweb_sources:
        logger.warning(
            "saml: 'saml' section is missing in cubicweb sources file")

    elif 'saml-metadata-uri' not in cubicweb_sources['saml']:
        logger.warning(
            "saml: 'saml-metadata-uri' option is missing")

    elif not cubicweb_sources['saml']['saml-metadata-uri']:
        logger.warning(
            "saml: 'saml-metadata-uri' option is empty")

    else:
        settings = config.get_settings()

        policy = SAMLAuthenticationPolicy(
            settings.get('cubicweb.auth.authtkt.session.secret'),
            settings.get('cubicweb.auth.authtkt.persistent.secret',
                         'notsosecret'))

        config.registry['cubicweb.authpolicy']._policies.append(policy)

        config.add_route('saml', '/saml')
        config.scan('cubicweb_saml.pconfig')
